import math
import random

import geomstats.backend as gs
from geomstats.geometry.hermitian import Hermitian, HermitianMetric
from tests.data_generation import (
    _ComplexRiemannianMetricTestData,
    _ComplexVectorSpaceTestData,
)

SQRT_2 = math.sqrt(2)
SQRT_5 = math.sqrt(5)


class HermitianTestData(_ComplexVectorSpaceTestData):
    Space = Hermitian

    n_list = random.sample(range(2, 5), 2)
    space_args_list = [(n,) for n in n_list]
    shape_list = [(n, n) for n in n_list]
    n_points_list = random.sample(range(2, 5), 2)
    n_vecs_list = random.sample(range(2, 5), 2)

    def belongs_test_data(self):
        smoke_data = [
            dict(dim=2, vec=gs.array([0.0, 1.0]), expected=True),
            dict(dim=2, vec=gs.array([1.0, 0.0, 1.0]), expected=False),
        ]
        return self.generate_tests(smoke_data)


class HermitianMetricTestData(_ComplexRiemannianMetricTestData):
    n_list = random.sample(range(2, 7), 5)

    shape_list = [(n,) for n in n_list]
    space_list = [Hermitian(n, equip=False) for n in n_list]
    metric_args_list = [{} for _ in shape_list]

    n_points_list = random.sample(range(1, 7), 5)
    n_tangent_vecs_list = random.sample(range(1, 7), 5)
    n_points_a_list = random.sample(range(1, 7), 5)
    n_points_b_list = [1]
    alpha_list = [1] * 5
    n_rungs_list = [1] * 5
    scheme_list = ["pole"] * 5

    Metric = HermitianMetric

    def exp_test_data(self):
        herm_2 = Hermitian(2)

        one_tangent_vec = gs.array([0.0 + 0.0j, 1.0 + 1.0j])
        one_base_point = gs.array([2.0 + 2.0j, 10.0 + 10.0j])
        n_tangent_vecs = gs.array(
            [
                [2.0 + 2.0j, 1.0 + 1.0j],
                [-2.0 - 2.0j, -4.0 - 4.0j],
                [-5.0 - 5.0j, 1.0 + 1.0j],
            ]
        )
        n_base_points = gs.array(
            [
                [2.0 + 2.0j, 10.0 + 10.0j],
                [8.0 + 8.0j, -1.0 - 1.0j],
                [-3.0 - 3.0j, 6.0 + 6.0j],
            ]
        )
        smoke_data = [
            dict(
                space=herm_2,
                tangent_vec=gs.array([0.0 + 0.0j, 1.0 + 1.0j]),
                base_point=gs.array([2.0 + 2.0j, 10.0 + 10.0j]),
                expected=gs.array([2.0 + 2.0j, 11.0 + 11.0j]),
            ),
            dict(
                space=herm_2,
                tangent_vec=one_tangent_vec,
                base_point=one_base_point,
                expected=one_tangent_vec + one_base_point,
            ),
            dict(
                space=herm_2,
                tangent_vec=one_tangent_vec,
                base_point=n_base_points,
                expected=one_tangent_vec + n_base_points,
            ),
            dict(
                space=herm_2,
                tangent_vec=n_tangent_vecs,
                base_point=one_base_point,
                expected=n_tangent_vecs + one_base_point,
            ),
            dict(
                space=herm_2,
                tangent_vec=n_tangent_vecs,
                base_point=n_base_points,
                expected=n_tangent_vecs + n_base_points,
            ),
        ]
        return self.generate_tests(smoke_data)

    def log_test_data(self):
        herm_2 = Hermitian(2, equip=False)
        one_p = gs.array([0.0 + 0.0j, 1.0 + 1.0j])
        one_bp = gs.array([2.0 + 2.0j, 10.0 + 10.0j])
        n_ps = gs.array(
            [
                [2.0 + 2.0j, 1.0 + 1.0j],
                [-2.0 - 2.0j, -4.0 - 4.0j],
                [-5.0 - 5.0, 1.0 + 1.0],
            ]
        )
        n_bps = gs.array(
            [
                [2.0 + 2.0, 10.0 + 10.0],
                [8.0 + 8.0, -1.0 - 1.0],
                [-3.0 - 3.0, 6.0 + 6.0],
            ]
        )
        smoke_data = [
            dict(
                space=herm_2,
                point=gs.array([2.0 + 2.0j, 10.0 + 10.0j]),
                base_point=gs.array([0.0 + 0.0j, 1.0 + 1.0j]),
                expected=gs.array([2.0 + 2.0j, 9.0 + 9.0j]),
            ),
            dict(space=herm_2, point=one_p, base_point=one_bp, expected=one_p - one_bp),
            dict(space=herm_2, point=one_p, base_point=n_bps, expected=one_p - n_bps),
            dict(space=herm_2, point=n_ps, base_point=one_bp, expected=n_ps - one_bp),
            dict(space=herm_2, point=n_ps, base_point=n_bps, expected=n_ps - n_bps),
        ]
        return self.generate_tests(smoke_data)

    def inner_product_test_data(self):
        herm_2 = Hermitian(2, equip=False)

        n_tangent_vecs_1 = gs.array([[2.0, 1.0], [-2.0, -4.0], [-5.0, 1.0]])
        n_tangent_vecs_2 = gs.array([[2.0, 10.0], [8.0, -1.0], [-3.0, 6.0]])
        tangent_vec_3 = gs.array([0.0, 1.0])
        tangent_vec_4 = gs.array([2.0, 10.0])
        smoke_data = [
            dict(
                space=herm_2,
                tangent_vec_a=n_tangent_vecs_1,
                tangent_vec_b=tangent_vec_4,
                expected=gs.array([14.0, -44.0, 0.0]),
            ),
            dict(
                space=herm_2,
                tangent_vec_a=tangent_vec_3,
                tangent_vec_b=n_tangent_vecs_2,
                expected=gs.array([10.0, -1.0, 6.0]),
            ),
            dict(
                space=herm_2,
                tangent_vec_a=n_tangent_vecs_1,
                tangent_vec_b=n_tangent_vecs_2,
                expected=gs.array([14.0, -12.0, 21.0]),
            ),
            dict(
                space=herm_2,
                tangent_vec_a=gs.array([0.0, 1.0]),
                tangent_vec_b=gs.array([2.0, 10.0]),
                expected=10.0,
            ),
        ]
        return self.generate_tests(smoke_data)

    def squared_norm_test_data(self):
        herm_2 = Hermitian(2, equip=False)

        smoke_data = [
            dict(
                space=herm_2,
                vec=gs.array([0.0, 1.0]),
                expected=1.0,
            ),
            dict(
                space=herm_2,
                vec=gs.array([[2.0, 1.0], [-2.0, -4.0], [-5.0, 1.0]]),
                expected=gs.array([5.0, 20.0, 26.0]),
            ),
        ]
        return self.generate_tests(smoke_data)

    def norm_test_data(self):
        smoke_data = [
            dict(
                space=Hermitian(2, equip=False), vec=gs.array([4.0, 3.0]), expected=5.0
            ),
            dict(
                space=Hermitian(4, equip=False),
                vec=gs.array([4.0, 3.0, 4.0, 3.0]),
                expected=5.0 * SQRT_2,
            ),
            dict(
                space=Hermitian(3, equip=False),
                vec=gs.array([[4.0, 3.0, 10.0], [3.0, 10.0, 4.0]]),
                expected=gs.array([5 * SQRT_5, 5 * SQRT_5]),
            ),
        ]
        return self.generate_tests(smoke_data)

    def metric_matrix_test_data(self):
        smoke_data = [
            dict(space=Hermitian(1, equip=False), expected=gs.eye(1)),
            dict(space=Hermitian(2, equip=False), expected=gs.eye(2)),
            dict(space=Hermitian(3, equip=False), expected=gs.eye(3)),
        ]
        return self.generate_tests(smoke_data)

    def squared_dist_test_data(self):
        one_point_a = gs.array([0.0, 1.0])
        one_point_b = gs.array([2.0, 10.0])
        n_points_a = gs.array([[2.0, 1.0], [-2.0, -4.0], [-5.0, 1.0]])
        n_points_b = gs.array([[2.0, 10.0], [8.0, -1.0], [-3.0, 6.0]])
        smoke_data = []
        smoke_data.append(
            dict(
                space=Hermitian(3, equip=False),
                point_a=one_point_a,
                point_b=n_points_b,
                expected=[85.0, 68.0, 34.0],
            )
        )
        smoke_data.append(
            dict(
                space=Hermitian(2, equip=False),
                point_a=one_point_a,
                point_b=one_point_b,
                expected=85.0,
            )
        )
        smoke_data.append(
            dict(
                space=Hermitian(2, equip=False),
                point_a=n_points_a,
                point_b=one_point_b,
                expected=[81.0, 212.0, 130.0],
            )
        )
        smoke_data.append(
            dict(
                space=Hermitian(2, equip=False),
                point_a=n_points_a,
                point_b=n_points_b,
                expected=[81.0, 109.0, 29.0],
            )
        )

        return self.generate_tests(smoke_data)

    def dist_test_data(self):
        herm_2 = Hermitian(2, equip=False)

        one_point_a = gs.array([0.0, 1.0])
        one_point_b = gs.array([2.0, 10.0])
        n_points_a = gs.array([[2.0, 1.0], [-2.0, -4.0], [-5.0, 1.0]])
        n_points_b = gs.array([[2.0, 10.0], [8.0, -1.0], [-3.0, 6.0]])
        smoke_data = []
        smoke_data.append(
            dict(
                space=herm_2,
                point_a=one_point_a,
                point_b=n_points_b,
                expected=gs.sqrt(gs.array([85.0, 68.0, 34.0])),
            )
        )
        smoke_data.append(
            dict(
                space=herm_2,
                point_a=one_point_a,
                point_b=one_point_b,
                expected=gs.sqrt(gs.array(85.0)),
            )
        )
        smoke_data.append(
            dict(
                space=herm_2,
                point_a=n_points_a,
                point_b=one_point_b,
                expected=gs.sqrt(gs.array([81.0, 212.0, 130.0])),
            )
        )
        smoke_data.append(
            dict(
                space=herm_2,
                point_a=n_points_a,
                point_b=n_points_b,
                expected=gs.sqrt(gs.array([81.0, 109.0, 29.0])),
            )
        )

        return self.generate_tests(smoke_data)
